/*****************************************************************************
 * File:	jugheadConf.c
 *
 * Author:	Rhett "Jonzy" Jones
 *		jonzy@cc.utah.edu
 *
 * Date:	June 6, 1994
 *
 * Modifed:	June 7, 1994 by Rhett "Jonzy" Jones.
 *		Forgot to add support for new-line '\n',
 *		carriage-return '\r', and tab '\t'.  Oops.
 *
 * Description:	Reads information from the jughead.conf file.
 *		The format of this file is such that any text following
 *		the pound sign "#" is a comment, blank lines are ignored,
 *		and contains a variable name followed by an assignment.
 *		A maximum of 1024 character per line is allowed, as
 *		well as embedded quotes, quoted and non-quoted strings.
 *		A sample file follows:
 *		# <- Start of a comment
 *		hostname = some.machine.name  # Not quoted
 *		hostname "some.machine.name"  # Does the same thing
 *		var = this one and that one   # var now gets [this]
 *		var = "This one and that one" # var gets [the quoted string]
 *		var "A \"quoted\" word"       # var gets [A "quoted" word]
 *
 * Notes:	When adding variables for use by ReadConfFile() you need
 *		to do the following steps:
 *		Add the #define variable_number
 *		Add the #define variable_name
 *		Add the declaration of the variables
 *		Add the else if (!strcmp(...)) in ParseLine()
 *		Add the case variable_number in ReadConfFile()
 *
 * Routines:	static int  ReadLine(FILE *fp,char *line,int lineSize);
 *		static int  ParseLine(char *line,char **varStr);
 *		       char *MakeStr(char *oldStr,char *newStr);
 *		       int  ReadConfFile(char *fileName);
 *
 * Bugs:	No known bugs.
 *
 * Copyright:	Copyright 1993, 1994, University of Utah Computer Center.
 *		This source may be freely distributed as long as this copyright
 * 		notice remains intact, and is in no way used for any monetary
 * 		gain, by any institution, business, person, or persons.
 *
 ****************************************************************************/

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <ctype.h>

#define BUFSIZE		1024

/* The variable-number to handle. */
#define HOSTNAME	1
#define JUGHEADHELP	2
#define ERRORHOST	3
#ifdef VERSION2
#define EMAIL		4
#define HOSTSTREE	5
#endif

/* The name of variable_name to handle. */
#define HOSTNAMESTR	"hostname"
#define JUGHEADHELPSTR	"jugheadhelp"
#define ERRORHOSTSTR	"errorhost"
#ifdef VERSION2
#define EMAILSTR	"email"
#define HOSTSTREESTR	"hoststree"
#endif

/* The declaration of the variable to handle.  All declared in "search.c". */
extern char	*hostname,
		*jugheadhelp,
		*errorhost;
#ifdef VERSION2
		*email,
		*hoststree;
#endif

/*****************************************************************************
 * ReadLine reads from the file 'fp' a maximum of 'lineSize' characters into
 * the string 'line' upto but not including end-of-file or end-of-line.
 * This routine returns the charcter read in.
 * Note:  If more them 'lineSize' character are found to exist on a line
 * a warning message is printed only once.
 ****************************************************************************/
static int ReadLine(fp,line,lineSize)
	FILE	*fp;		/* Pointer to the file we are reading. */
	char	*line;		/* Buffer for the line we read. */
	int	lineSize;	/* The size of our line buffer. */
{	int	i,c;		/* A loop counter and a character. */

	/* Read the charcters up to the limit, eof-of-file or end-of-line. */
	for (i = 0; i < lineSize && (c = fgetc(fp)) != EOF && c != '\n'; i++)
		line[i] = (char)c;
	line[i] = '\0';

	if (i >= lineSize)	/* Yipes, tell the user their file is out to lunch. */
		{
		static int	messagePosted = 0;	/* Have we told them yet? */

		if (!messagePosted)
			{
			messagePosted = 1;
			(void)fprintf(stderr,"warning: configuration file cannot contain more than %d characters per line.\n",lineSize + 1);
			(void)fprintf(stderr,"         All charcters beyond this limit are discarded.\n");
			}
		while ((c = fgetc(fp)) != EOF && c != '\n');
		}

	return(c);

}	/* ReadLine */

/*****************************************************************************
 * ParseLine returns what type of variable we have encountered within the
 * jughead.conf file.  This routine assigns to 'varStr' the contents of what
 * the encountered variable should get assigned, as well as takes care of
 * embedded quotes.  The "Get the variable contents" is kind of nitty gritty
 * but it handles quoted strings, embedded quotes, and non-quoted strings.
 ****************************************************************************/
static int ParseLine(line,varStr)
	char	*line,		/* The line we are parsing. */
		**varStr;	/* The string to assign the variable. */
{	int	varName = 0,	/* What variable did we encounter? */
		i, j,		/* Loop and index counters. */
		done,		/* Are we done with string contents? */
		escape,		/* Have we encountered an escape character? */
		quote;		/* Have we seen a quote? */
	char	*s;		/* Place holder into 'line'. */

	/* Skip any leading whitespace. */
	while (isspace(*line))
		line++;

	if (*line == '#')	/* Comment line so bail out. */
		return(0);

	for (s = line; *line && !isspace(*line) && *line != '=' && *line != '#'; line++)
		/* Get the name of the variable. */;

	/* See if we got the variable name. */
	if (isspace(*line) || *line == '=')
		{
		*line++ = '\0';	/* 's' now contains the variable name. */

		/* Find out which variable we are looking at. */
		if (!strcmp(s,HOSTNAMESTR))
			varName = HOSTNAME;
		else if (!strcmp(s,JUGHEADHELPSTR))
			varName = JUGHEADHELP;
		else if (!strcmp(s,ERRORHOSTSTR))
			varName = ERRORHOST;
#ifdef VERSION2
		else if (!strcmp(s,EMAILSTR))
			varName = EMAIL;
		else if (!strcmp(s,HOSTSTREESTR))
			varName = HOSTSTREE;
#endif
		else	/* Who the hell knows? */
			return(0);

		/* Skip any white space or equal signs. */
		while (isspace(*line) || *line == '=')
			line++;

		/* Get the variable contents. */
		for (done = escape = quote = i = j = 0, s = line; !done && line[i] && line[i] != '#'; i++)
			{
			if (line[i] == '\\')
				{
				escape++;
				j++;
				}
			else if (!escape && !quote && isspace(line[i]))	
				{
				done++;
				j++;
				}
			else if (line[i] == '\"')
				if (escape)
					{
					escape = 0;
					line[i - j] = line[i];
					}
				else
					{
					if (quote)
						done++;
					else
						quote++;
					j++;
					}
			else if (escape && (line[i] == 'n' || line[i] == 'r' || line[i] == 't'))
				{
				escape = 0;
				switch (line[i])
					{
					case 'n': line[i - j] = '\n';	break;
					case 'r': line[i - j] = '\r';	break;
					case 't': line[i - j] = '\t';	break;
					}
				}
			else
				line[i - j] = line[i];
			}
		line[i - j] = '\0';	/* 's' now contains the variable contents. */

		*varStr = s;
		}

	return(varName);

}	/* ParseLine */

/*****************************************************************************
 * MakeStr returns a copy of 'newStr' acquired in malloc space.  This
 * routine also frees up any memory occupied by 'oldStr' prior to the copy.
 ****************************************************************************/
char *MakeStr(oldStr,newStr)
	char	*oldStr,	/* The string to copy into. */
		*newStr;	/* The string we are copying. */
{
	if (oldStr)	/* Free up the memory. */
		(void)free(oldStr);

	/* Get memory and copy the sucker. */
	if (oldStr = (char *)malloc((unsigned)strlen(newStr) * sizeof(char) + 1))
		(void)strcpy(oldStr,newStr);

	return(oldStr);

}	/* MakeStr */

/*****************************************************************************
 * ReadConfFile reads from the file 'fileName', which is the jughead.conf
 * file, and initializes the variables stored within the file.  This routine
 * returns true if the file exists and false otherwise.
 ****************************************************************************/
int ReadConfFile(fileName)
	char	*fileName;	/* Name of the jughead.conf file. */
{	FILE	*fp;		/* Pointer to the conf file. */
	char	line[BUFSIZE],
		*varStr;

	if (fp = fopen(fileName,"r"))
		{
		while (ReadLine(fp,line,BUFSIZE - 1) != EOF)
			switch (ParseLine(line,&varStr))
				{
				case HOSTNAME:	  hostname    = MakeStr(hostname,varStr);	break;
				case JUGHEADHELP: jugheadhelp = MakeStr(jugheadhelp,varStr);	break;
				case ERRORHOST:	  errorhost   = MakeStr(errorhost,varStr);	break;
#ifdef VERSION2
				case EMAIL:	  email       = MakeStr(email,varStr);		break;
				case HOSTSTREE:	  hoststree   = MakeStr(hoststree,varStr);	break;
#endif
				default:							break;
				}
		(void)fclose(fp);

		return(1);
		}

	return(0);

}	/* ReadConfFile */

